@@ -37,6 +37,7 @@
 #include "gedit-utils.h"
 #include "gedit-multi-notebook.h"
 #include "gedit-notebook.h"
+#include "gedit-cell-renderer-button.h"
 
 #include <glib/gi18n.h>
 
@@ -52,12 +53,14 @@ struct _GeditDocumentsPanelPrivate
 	GtkWidget          *treeview;
 	GtkTreeModel       *model;
 
+	guint               refresh_idle_id;
+
 	guint               adding_tab : 1;
 	guint               is_reodering : 1;
 	guint               setting_active_notebook : 1;
 };
 
-G_DEFINE_TYPE(GeditDocumentsPanel, gedit_documents_panel, GTK_TYPE_VBOX)
+G_DEFINE_TYPE(GeditDocumentsPanel, gedit_documents_panel, GTK_TYPE_BOX)
 
 enum
 {
@@ -160,85 +163,84 @@ get_iter_from_tab (GeditDocumentsPanel *panel,
 
 	gedit_debug (DEBUG_PANEL);
 
-	g_assert (notebook != NULL || tab != NULL);
+	g_assert (notebook != NULL);
+	/* tab may be NULL if just the notebook is specified */
 	g_assert (tab_iter != NULL);
-	g_assert ((tab == NULL && tab_iter == NULL) || (tab != NULL && tab_iter != NULL));
 
-	success = FALSE;
 	search_notebook = (gedit_multi_notebook_get_n_notebooks (panel->priv->mnb) > 1);
 
-	/* We never ever ask for a notebook or tab if it does not already exist in the tree */
-	g_assert (gtk_tree_model_get_iter_first (panel->priv->model, &parent));
+	if (!gtk_tree_model_get_iter_first (panel->priv->model, &parent))
+		return FALSE;
 
+	success = FALSE;
 	do
 	{
+		GtkTreeIter iter;
+
 		if (search_notebook)
 		{
 			GeditNotebook *current_notebook;
-			gboolean is_cur;
 
 			gtk_tree_model_get (panel->priv->model,
 					    &parent,
 					    NOTEBOOK_COLUMN, &current_notebook,
 					    -1);
 
-			is_cur = (current_notebook == notebook);
-
-			g_object_unref (current_notebook);
-
-			if (is_cur)
+			if (current_notebook != NULL)
 			{
-				success = TRUE;
+				gboolean is_cur;
+				is_cur = (current_notebook == notebook);
 
-				if (tab == NULL)
+				g_object_unref (current_notebook);
+
+				if (is_cur)
 				{
-					break;
+					/* notebook found */
+					search_notebook = FALSE;
+					gtk_tree_model_iter_children (panel->priv->model, &iter, &parent);
 				}
 			}
 		}
-		else if (tab == NULL)
+		else
 		{
-			success = TRUE;
-
-			break;
+			iter = parent;
 		}
 
-		if (success || !search_notebook)
+		if (!search_notebook)
 		{
-			GeditTab *current_tab;
-			GtkTreeIter iter;
+			if (tab == NULL)
+			{
+				success = TRUE;
+				break;
+			}
 
-			if (search_notebook)
-			{
-				g_assert (gtk_tree_model_iter_children (panel->priv->model,
-									&iter, &parent));
-			}
-			else
-			{
-				iter = parent;
-			}
+			g_assert (gtk_tree_store_iter_is_valid (GTK_TREE_STORE (panel->priv->model),
+								&iter));
 
 			do
 			{
-				gboolean is_cur;
+				GeditTab *current_tab;
 
 				gtk_tree_model_get (panel->priv->model,
 						    &iter,
 						    TAB_COLUMN, &current_tab,
 						    -1);
-				g_assert (current_tab != NULL);
 
-				is_cur = (current_tab == tab);
-
-				g_object_unref (current_tab);
-
-				if (is_cur)
+				if (current_tab != NULL)
 				{
-					*tab_iter = iter;
-					success = TRUE;
+					gboolean is_cur;
 
-					/* break 2; */
-					goto out;
+					is_cur = (current_tab == tab);
+					g_object_unref (current_tab);
+
+					if (is_cur)
+					{
+						*tab_iter = iter;
+						success = TRUE;
+
+						/* break 2; */
+						goto out;
+					}
 				}
 			} while (gtk_tree_model_iter_next (panel->priv->model, &iter));
 
@@ -249,15 +251,6 @@ get_iter_from_tab (GeditDocumentsPanel *panel,
 
 out:
 
-	/* In the current code if we are not successful then something is wrong */
-	g_assert (success);
-
-	if (tab_iter != NULL)
-	{
-		g_assert (gtk_tree_store_iter_is_valid (GTK_TREE_STORE (panel->priv->model),
-							tab_iter));
-	}
-
 	return success;
 }
 
@@ -296,8 +289,8 @@ select_active_tab (GeditDocumentsPanel *panel)
 	{
 		GtkTreeIter iter;
 
-		get_iter_from_tab (panel, notebook, tab, &iter);
-		select_iter (panel, &iter);
+		if (get_iter_from_tab (panel, notebook, tab, &iter))
+			select_iter (panel, &iter);
 	}
 }
 
@@ -316,10 +309,8 @@ multi_notebook_tab_switched (GeditMultiNotebook  *mnb,
 	{
 		GtkTreeIter iter;
 
-		get_iter_from_tab (panel, new_notebook, new_tab, &iter);
-
-		if (gtk_tree_store_iter_is_valid (GTK_TREE_STORE (panel->priv->model),
-						  &iter))
+		if (get_iter_from_tab (panel, new_notebook, new_tab, &iter) &&
+		    gtk_tree_store_iter_is_valid (GTK_TREE_STORE (panel->priv->model), &iter))
 		{
 			select_iter (panel, &iter);
 		}
@@ -416,8 +407,8 @@ refresh_notebook_foreach (GeditNotebook       *notebook,
 	}
 }
 
-static void
-refresh_list (GeditDocumentsPanel *panel)
+static gboolean
+refresh_list_idle (GeditDocumentsPanel *panel)
 {
 	gedit_debug (DEBUG_PANEL);
 
@@ -432,6 +423,22 @@ refresh_list (GeditDocumentsPanel *panel)
 	gtk_tree_view_expand_all (GTK_TREE_VIEW (panel->priv->treeview));
 
 	select_active_tab (panel);
+
+	panel->priv->refresh_idle_id = 0;
+
+	return FALSE;
+}
+
+static void
+refresh_list (GeditDocumentsPanel *panel)
+{
+	/* refresh in an idle so that when adding/removing many tabs
+	 * the model is repopulated just once */
+	if (panel->priv->refresh_idle_id == 0)
+	{
+		panel->priv->refresh_idle_id = gdk_threads_add_idle ((GSourceFunc) refresh_list_idle,
+		                                                     panel);
+	}
 }
 
 static void
@@ -452,29 +459,32 @@ sync_name_and_icon (GeditTab            *tab,
 		    GParamSpec          *pspec,
 		    GeditDocumentsPanel *panel)
 {
-	GdkPixbuf *pixbuf;
-	gchar *name;
 	GtkTreeIter iter;
 
 	gedit_debug (DEBUG_PANEL);
 
-	get_iter_from_tab (panel,
-			   gedit_multi_notebook_get_active_notebook (panel->priv->mnb),
-			   tab, &iter);
-
-	name = tab_get_name (tab);
-	pixbuf = _gedit_tab_get_icon (tab);
-
-	gtk_tree_store_set (GTK_TREE_STORE (panel->priv->model),
-			    &iter,
-			    PIXBUF_COLUMN, pixbuf,
-			    NAME_COLUMN, name,
-			    -1);
-
-	g_free (name);
-	if (pixbuf != NULL)
+	if (get_iter_from_tab (panel,
+			       gedit_multi_notebook_get_active_notebook (panel->priv->mnb),
+			       tab,
+			       &iter))
 	{
-		g_object_unref (pixbuf);
+		gchar *name;
+		GdkPixbuf *pixbuf;
+
+		name = tab_get_name (tab);
+		pixbuf = _gedit_tab_get_icon (tab);
+
+		gtk_tree_store_set (GTK_TREE_STORE (panel->priv->model),
+				    &iter,
+				    PIXBUF_COLUMN, pixbuf,
+				    NAME_COLUMN, name,
+				    -1);
+
+		g_free (name);
+		if (pixbuf != NULL)
+		{
+			g_object_unref (pixbuf);
+		}
 	}
 }
 
@@ -681,6 +691,12 @@ gedit_documents_panel_dispose (GObject *object)
 
 	gedit_debug (DEBUG_PANEL);
 
+	if (panel->priv->refresh_idle_id != 0)
+	{
+		g_source_remove (panel->priv->refresh_idle_id);
+		panel->priv->refresh_idle_id = 0;
+	}
+
 	if (panel->priv->window != NULL)
 	{
 		g_object_unref (panel->priv->window);
@@ -740,7 +756,7 @@ menu_position (GtkMenu             *menu,
 {
 	GtkTreePath *path;
 	GdkRectangle rect;
-	gint wx, wy;
+	gint wy;
 	GtkRequisition requisition;
 	GtkWidget *w;
 	GtkAllocation allocation;
@@ -756,7 +772,6 @@ menu_position (GtkMenu             *menu,
 				     NULL,
 				     &rect);
 
-	wx = rect.x;
 	wy = rect.y;
 
 	gdk_window_get_origin (gtk_widget_get_window (w), x, y);
@@ -1010,6 +1025,40 @@ treeview_row_inserted (GtkTreeModel        *tree_model,
 }
 */
 
+static void
+close_button_clicked (GtkCellRenderer     *cell,
+                      const gchar         *path,
+                      GeditDocumentsPanel *panel)
+{
+        GtkTreeIter iter;
+        GeditTab *tab;
+        GeditNotebook *notebook;
+
+        if (!gtk_tree_model_get_iter_from_string (panel->priv->model,
+                                                  &iter, path))
+        {
+                return;
+        }
+
+        gtk_tree_model_get (panel->priv->model,
+		            &iter,
+		            NOTEBOOK_COLUMN, &notebook,
+		            TAB_COLUMN, &tab,
+		            -1);
+
+	if (tab == NULL)
+	{
+		gedit_notebook_remove_all_tabs (notebook);
+	}
+	else
+	{
+		gedit_notebook_remove_tab (notebook, tab);
+		g_object_unref (tab);
+	}
+
+	g_object_unref (notebook);
+}
+
 static void
 pixbuf_data_func (GtkTreeViewColumn   *column,
 		  GtkCellRenderer     *cell,
@@ -1047,6 +1096,9 @@ gedit_documents_panel_init (GeditDocumentsPanel *panel)
 	panel->priv->adding_tab = FALSE;
 	panel->priv->is_reodering = FALSE;
 
+	gtk_orientable_set_orientation (GTK_ORIENTABLE (panel),
+	                                GTK_ORIENTATION_VERTICAL);
+
 	/* Create the scrolled window */
 	sw = gtk_scrolled_window_new (NULL, NULL);
 
@@ -1102,6 +1154,14 @@ gedit_documents_panel_init (GeditDocumentsPanel *panel)
 	gtk_tree_view_append_column (GTK_TREE_VIEW (panel->priv->treeview),
 				     column);
 
+	cell = gedit_cell_renderer_button_new ();
+	g_object_set (cell, "stock-id", GTK_STOCK_CLOSE, NULL);
+	gtk_tree_view_column_pack_end (column, cell, FALSE);
+	g_signal_connect (cell,
+	                  "clicked",
+	                  G_CALLBACK (close_button_clicked),
+	                  panel);
+
 	selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (panel->priv->treeview));
 
 	gtk_tree_selection_set_mode (selection, GTK_SELECTION_SINGLE);
@@ -1138,4 +1198,4 @@ gedit_documents_panel_new (GeditWindow *window)
 			     NULL);
 }
 
-/* ex:ts=8:noet: */
+/* ex:set ts=8 noet: */
